using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;

namespace TankLevelEditor
{
    public delegate void ItemCallback(object item);

    public partial class LevelEditor : UserControl
    {
        private enum MouseMode
        {
            None,
            DraggingItemPoint,
        }

        private int AreaSize
        {
            get { return Math.Max(Math.Min(ClientSize.Width, ClientSize.Height), 1); } 
        }

        private Level m_level;
        public Level EditLevel
        {
            get { return m_level; }
            set
            {
                if (m_level == value)
                    return;

                ActiveItem = null;

                m_level = value;
                if (LevelChangedEvent != null)
                    LevelChangedEvent(m_level);
                Invalidate();
            }
        }

        private LevelItem m_active_item;
        public LevelItem ActiveItem
        {
            get { return m_active_item; }
            set
            {
                if (m_active_item == value)
                    return;
                m_active_item = value;
                if (ActiveItemChangedEvent != null)
                    ActiveItemChangedEvent(m_active_item);
                Invalidate();
            }
        }

        public event ItemCallback LevelChangedEvent;
        public event ItemCallback ActiveItemChangedEvent;

        private MouseMode m_mouse_mode = MouseMode.None;
        private int m_mouse_x, m_mouse_y;
        private int m_mouse_drag_idx;

        public LevelEditor()
        {
            InitializeComponent();
        }

        private void LevelEditor_Paint(object sender, PaintEventArgs e)
        {
            Graphics g = e.Graphics;

            g.Clear(Color.Gray);

            // Background fill
            int area_size = AreaSize;
            if (EditLevel == null)
            {
                g.FillRectangle(Brushes.DarkGray, 0, 0, area_size, area_size);
                return;
            }
            g.FillRectangle(Brushes.Black, 0, 0, area_size, area_size);
            
            // Draw all tanks
            foreach (Tank tank in EditLevel.tanks)
            {
                bool selected = (tank == ActiveItem);
                PointF scr_pt = MapToScreen(tank.pos);
                float scr_size = MapToScreen(0.5f);

                Brush tank_brush = Brushes.Gray;
                if (tank.FixedPosition)
                    tank_brush = Brushes.Red;
                if (selected)
                    tank_brush = Brushes.Yellow;

                Pen tank_pen = new Pen(tank_brush);

                g.DrawEllipse(tank_pen, scr_pt.X - scr_size, scr_pt.Y - scr_size, scr_size * 2, scr_size * 2);
                g.FillRectangle(tank_brush, scr_pt.X - 2, scr_pt.Y - 2, 4, 4);
            }

            // Draw all walls
            foreach (Wall wall in EditLevel.walls)
            {
                bool selected = (wall == ActiveItem);
                PointF pt1 = MapToScreen(wall.pt1);
                PointF pt2 = MapToScreen(wall.pt2);
                float thickness = MapToScreen(wall.Thickness) * 0.5f;

                float nx = -(pt2.Y - pt1.Y);
                float ny = (pt2.X - pt1.X);
                float nlen = (float)Math.Sqrt(nx * nx + ny * ny);
                if (nlen > 0.01f)
                {
                    nx /= nlen;
                    ny /= nlen;
                }
                else
                {
                    nx = 0;
                    ny = 0;
                }

                Brush wall_brush = Brushes.Gray;
                if (wall.Type == WallType.HoleShallow || wall.Type == WallType.HoleDeep)
                    wall_brush = Brushes.Blue;
                if (selected)
                    wall_brush = Brushes.Yellow;

                Pen wall_pen = new Pen(wall_brush);

                // Direct line
                //g.DrawLine(wall_pen, pt1, pt2);

                // Outline
                g.DrawLine(wall_pen,
                    pt1.X - nx * thickness, pt1.Y - ny * thickness,
                    pt1.X + nx * thickness, pt1.Y + ny * thickness);
                g.DrawLine(wall_pen,
                    pt2.X - nx * thickness, pt2.Y - ny * thickness,
                    pt2.X + nx * thickness, pt2.Y + ny * thickness);
                g.DrawLine(wall_pen,
                    pt1.X - nx * thickness, pt1.Y - ny * thickness,
                    pt2.X - nx * thickness, pt2.Y - ny * thickness);
                g.DrawLine(wall_pen,
                    pt1.X + nx * thickness, pt1.Y + ny * thickness,
                    pt2.X + nx * thickness, pt2.Y + ny * thickness);

                // Endpoints
                g.FillRectangle(wall_brush, pt1.X - 2, pt1.Y - 2, 4, 4);
                g.FillRectangle(wall_brush, pt2.X - 2, pt2.Y - 2, 4, 4);
            }
        }

        private void LevelEditor_Resize(object sender, EventArgs e)
        {
            Invalidate();
        }

        // -1 for "no touch"
        private float MouseDistToItem(LevelItem item, int mx, int my, out int pt_idx)
        {
            pt_idx = -1;

            for (int i = 0; i < item.PointCount; ++i)
            {
                PointF scr_pt = MapToScreen(item.GetPoint(i));
                float dx = scr_pt.X - (float)mx;
                float dy = scr_pt.Y - (float)my;
                float dist = (float)Math.Sqrt(dx * dx + dy * dy);

                if (dist <= 10)
                {
                    pt_idx = i;
                    return dist;
                }
            }

            return -1;
        }

        private float MapToScreen(float xy)
        {
            return (xy / 10.0f) * (float)AreaSize;
        }

        private PointF MapToScreen(PointF pt)
        {
            return new PointF(MapToScreen(pt.X), MapToScreen(pt.Y));
        }

        private float ScreenToMap(float xy)
        {
            return (xy / (float)AreaSize) * 10.0f;
        }

        private PointF ScreenToMap(PointF pt)
        {
            return new PointF(ScreenToMap(pt.X), ScreenToMap(pt.Y));
        }

        private void LevelEditor_MouseDown(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
            {
                m_mouse_x = e.X;
                m_mouse_y = e.Y;

                // Clicked on a selected item?
                if (ActiveItem != null)
                {
                    int pt_idx;
                    if (MouseDistToItem(ActiveItem, e.X, e.Y, out pt_idx) >= 0.0f)
                    {
                        m_mouse_mode = MouseMode.DraggingItemPoint;
                        m_mouse_drag_idx = pt_idx;
                        Invalidate();
                        return;
                    }
                }

                // Picked a new item/deselected
                LevelItem new_item = null;
                float last_dist = 0;

                // Check all tanks
                foreach (Tank tank in EditLevel.tanks)
                {
                    int pt_idx;
                    float dist = MouseDistToItem(tank, e.X, e.Y, out pt_idx);
                    if (dist < 0)
                        continue;

                    if (new_item == null || dist < last_dist)
                    {
                        new_item = tank;
                        last_dist = dist;
                    }
                }

                // Check all walls
                foreach (Wall wall in EditLevel.walls)
                {
                    int pt_idx;
                    float dist = MouseDistToItem(wall, e.X, e.Y, out pt_idx);
                    if (dist < 0)
                        continue;

                    if (new_item == null || dist < last_dist)
                    {
                        new_item = wall;
                        last_dist = dist;
                    }
                }

                ActiveItem = new_item;
            }
        }

        private void LevelEditor_MouseUp(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
            {
                m_mouse_mode = MouseMode.None;
            }
        }

        private void LevelEditor_MouseMove(object sender, MouseEventArgs e)
        {
            int dx = e.X - m_mouse_x;
            int dy = e.Y - m_mouse_y;
            m_mouse_x = e.X;
            m_mouse_y = e.Y;

            if (m_mouse_mode == MouseMode.DraggingItemPoint &&
                ActiveItem != null)
            {
                PointF pt = ActiveItem.GetPoint(m_mouse_drag_idx);
                pt.X += ScreenToMap(dx);
                pt.Y += ScreenToMap(dy);

                if (pt.X < 0) pt.X = 0;
                if (pt.Y < 0) pt.Y = 0;
                if (pt.X > 10) pt.X = 10;
                if (pt.Y > 10) pt.Y = 10;

                ActiveItem.SetPoint(m_mouse_drag_idx, pt);
                Invalidate();
            }
        }
    }
}
